import re
import time
from typing import Any, Sequence, Union

import numpy as np
from pyvisa import constants, errors
from qcodes.instrument import ChannelList, InstrumentChannel, VisaInstrument
from qcodes.parameters import (
    Parameter,
    ParameterBase,
    ParameterWithSetpoints,
    create_on_off_val_mapping,
)
from qcodes.validators import Arrays, Bool, Enum, Ints, Numbers


class PNAAxisParameter(Parameter):
    def __init__(
        self,
        startparam: Parameter,
        stopparam: Parameter,
        pointsparam: Parameter,
        **kwargs: Any,
    ):
        """
        Axis parameter for traces from the PNA
        """
        super().__init__(**kwargs)

        self._startparam = startparam
        self._stopparam = stopparam
        self._pointsparam = pointsparam

    def get_raw(self) -> np.ndarray:
        """
        Return the axis values, with values retrieved from the parent instrument
        """
        return np.linspace(
            self._startparam(), self._stopparam(), self._pointsparam()
        )


class PNALogAxisParamter(PNAAxisParameter):
    def get_raw(self) -> np.ndarray:
        """
        Return the axis values on a log scale, with values retrieved from
        the parent instrument
        """
        return np.geomspace(
            self._startparam(), self._stopparam(), self._pointsparam()
        )


class PNATimeAxisParameter(PNAAxisParameter):
    def get_raw(self) -> np.ndarray:
        """
        Return the axis values on a time scale, with values retrieved from
        the parent instrument
        """
        return np.linspace(0, self._stopparam(), self._pointsparam())


class PNATimeDomainAxisParameter(PNAAxisParameter):
    def get_raw(self) -> np.ndarray:
        """
        Return the axis values on a time scale, with values retrieved from
        the parent instrument
        """
        return np.linspace(
            self._startparam(), self._stopparam(), self._pointsparam()
        )


class FormattedSweep(ParameterWithSetpoints):
    """
    Mag will run a sweep, including averaging, before returning data.
    As such, wait time in a loop is not needed.
    """

    def __init__(
        self,
        name: str,
        instrument: "PNABase",
        sweep_format: str,
        label: str,
        unit: str,
        memory: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            name, instrument=instrument, label=label, unit=unit, **kwargs
        )
        self.sweep_format = sweep_format
        self.memory = memory

    @property
    def setpoints(self) -> Sequence[ParameterBase]:
        """
        Overwrite setpoint parameter to ask the PNA what type of sweep
        """
        if self.instrument is None:
            raise RuntimeError(
                "Cannot return setpoints if not attached " "to instrument"
            )
        root_instrument: "PNABase" = self.root_instrument  # type: ignore[assignment]
        sweep_type = root_instrument.sweep_type()
        try:
            time_domain_enabled = root_instrument.time_domain_enabled()
        except:
            time_domain_enabled = 0
        if sweep_type == "LIN" and time_domain_enabled == 0:
            return (root_instrument.frequency_axis,)
        elif sweep_type == "LIN" and time_domain_enabled == 1:
            return (root_instrument.time_domain_axis,)
        elif sweep_type == "LOG":
            return (root_instrument.frequency_log_axis,)
        elif sweep_type == "CW":
            return (root_instrument.time_axis,)
        else:
            raise NotImplementedError(
                f"Axis for type {sweep_type} not implemented yet"
            )

    @setpoints.setter
    def setpoints(self, val: Any) -> None:
        """
        Stub to allow initialization. Ignore any set attempts on setpoint as we
        figure it out on the fly.
        """
        return

    def get_raw(self) -> Sequence[float]:
        if self.instrument is None:
            raise RuntimeError("Cannot get data without instrument")
        root_instr = self.instrument.root_instrument
        # Check if we should run a new sweep
        if root_instr.auto_sweep():
            prev_mode = self.instrument.run_sweep()
        # Ask for data, setting the format to the requested form
        self.instrument.format(self.sweep_format)
        data = root_instr.visa_handle.query_binary_values(
            "CALC:DATA? FDATA", datatype="f", is_big_endian=True
        )
        data = np.array(data)
        # Restore previous state if it was changed
        if root_instr.auto_sweep():
            root_instr.sweep_mode(prev_mode)

        return data


class PNAPort(InstrumentChannel):
    """
    Allow operations on individual PNA ports.
    Note: This can be expanded to include a large number of extra parameters...
    """

    def __init__(
        self,
        parent: "PNABase",
        name: str,
        port: int,
        min_power: Union[int, float],
        max_power: Union[int, float],
        **kwargs: Any,
    ) -> None:
        super().__init__(parent, name, **kwargs)

        self.port = int(port)
        if self.port < 1 or self.port > 4:
            raise ValueError("Port must be between 1 and 4.")

        pow_cmd = f"SOUR:POW{self.port}"
        self.add_parameter(
            "source_power",
            label="power",
            unit="dBm",
            get_cmd=f"{pow_cmd}?",
            set_cmd=f"{pow_cmd} {{}}",
            get_parser=float,
            vals=Numbers(min_value=min_power, max_value=max_power),
        )

    def _set_power_limits(
        self, min_power: Union[int, float], max_power: Union[int, float]
    ) -> None:
        """
        Set port power limits
        """
        self.source_power.vals = Numbers(
            min_value=min_power, max_value=max_power
        )


class PNATrace(InstrumentChannel):
    """
    Allow operations on individual PNA traces.
    """

    def __init__(
        self,
        parent: "PNABase",
        name: str,
        trace_name: str,
        trace_num: int,
        **kwargs: Any,
    ) -> None:
        super().__init__(parent, name, **kwargs)
        self.trace_name = trace_name
        self.trace_num = trace_num

        # Name of parameter (i.e. S11, S21 ...)
        self.add_parameter(
            "trace",
            label="Trace",
            get_cmd=self._Sparam,
            set_cmd=self._set_Sparam,
        )
        # Format
        # Note: Currently parameters that return complex values are not
        # supported as there isn't really a good way of saving them into the
        # dataset
        self.add_parameter(
            "format",
            label="Format",
            get_cmd="CALC:FORM?",
            set_cmd="CALC:FORM {}",
            vals=Enum("MLIN", "MLOG", "PHAS", "UPH", "IMAG", "REAL", "POLAR"),
        )

        # And a list of individual formats
        self.add_parameter(
            "magnitude",
            sweep_format="MLOG",
            label="Magnitude",
            unit="dB",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "linear_magnitude",
            sweep_format="MLIN",
            label="Magnitude",
            unit="",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "phase",
            sweep_format="PHAS",
            label="Phase",
            unit="deg",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "unwrapped_phase",
            sweep_format="UPH",
            label="Phase",
            unit="deg",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "group_delay",
            sweep_format="GDEL",
            label="Group Delay",
            unit="s",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "real",
            sweep_format="REAL",
            label="Real",
            unit="",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "imaginary",
            sweep_format="IMAG",
            label="Imaginary",
            unit="",
            parameter_class=FormattedSweep,
            vals=Arrays(shape=(self.parent.points,)),
        )
        self.add_parameter(
            "polar",
            sweep_format="POLAR",
            label="Polar",
            unit="",
            parameter_class=FormattedSweep,
            get_parser=self._parse_polar_data,
            vals=Arrays(shape=(self.parent.points,), valid_types=(complex,)),
        )

    @staticmethod
    def _parse_polar_data(data: Sequence[float]) -> np.ndarray:
        """
        Parse the 2*n-length flat array coming from the instrument
        and convert to n-length array of complex numbers
        """
        pairs = np.array([data[::2], data[1::2]]).T
        return np.array([complex(*pair) for pair in pairs])

    def run_sweep(self) -> str:
        """
        Run a set of sweeps on the network analyzer.
        Note that this will run all traces on the current channel.
        """
        root_instr = self.root_instrument
        # Store previous mode
        prev_mode = root_instr.sweep_mode()
        # Take instrument out of continuous mode, and send triggers equal to
        # the number of averages
        if root_instr.averages_enabled():
            avg = root_instr.averages()
            root_instr.reset_averages()
            root_instr.group_trigger_count(avg)
            root_instr.sweep_mode("GRO")
        else:
            root_instr.sweep_mode("SING")

        # Once the sweep mode is in hold, we know we're done
        try:
            while root_instr.sweep_mode() != "HOLD":
                time.sleep(0.1)
        except KeyboardInterrupt:
            # If the user aborts because (s)he is stuck in the infinite loop
            # mentioned above, provide a hint of what can be wrong.
            msg = "User abort detected. "
            source = root_instr.trigger_source()
            if source == "MAN":
                msg += (
                    "The trigger source is manual. Are you sure this is "
                    "correct? Please set the correct source with the "
                    "'trigger_source' parameter"
                )
            elif source == "EXT":
                msg += (
                    "The trigger source is external. Is the trigger "
                    "source functional?"
                )
            self.log.warning(msg)
            raise

        # Return previous mode, incase we want to restore this
        return prev_mode

    def write(self, cmd: str) -> None:
        """
        Select correct trace before querying
        """
        self.root_instrument.active_trace(self.trace_num)
        super().write(cmd)

    def ask(self, cmd: str) -> str:
        """
        Select correct trace before querying
        """
        self.root_instrument.active_trace(self.trace_num)
        return super().ask(cmd)

    def _Sparam(self) -> str:
        """
        Extrace S_parameter from returned PNA format
        """
        paramspec = self.root_instrument.get_trace_catalog()
        specs = paramspec.split(",")
        for spec_ind in range(len(specs) // 2):
            name, param = specs[spec_ind * 2 : (spec_ind + 1) * 2]
            if name == self.trace_name:
                return param
        raise RuntimeError("Can't find selected trace on the PNA")

    def _set_Sparam(self, val: str) -> None:
        """
        Set an S-parameter, in the format S<a><b>, where a and b
        can range from 1-2
        """
        if not re.match("S[1-2][1-2]", val):
            raise ValueError("Invalid S parameter spec")
        self.write(f'CALC:PAR:MOD:EXT "{val}"')


class PNABase(VisaInstrument):
    """
    Base qcodes driver for Agilent/Keysight series PNAs
    http://na.support.keysight.com/pna/help/latest/Programming/GP-IB_Command_Finder/SCPI_Command_Tree.htm
    Note: Currently this driver only expects a single channel on the PNA. We
        can handle multiple traces, but using traces across multiple channels
        may have unexpected results.
    """

    def __init__(
        self,
        name: str,
        address: str,
        # Set frequency ranges
        min_freq: Union[int, float],
        max_freq: Union[int, float],
        # Set power ranges
        min_power: Union[int, float],
        max_power: Union[int, float],
        nports: int,  # Number of ports on the PNA
        **kwargs: Any,
    ) -> None:
        super().__init__(name, address, terminator="\n", **kwargs)
        self.min_freq = min_freq
        self.max_freq = max_freq

        self.log.info(
            "Initializing %s with power range %r-%r, freq range %r-%r.",
            name,
            min_power,
            max_power,
            min_freq,
            max_freq,
        )

        # Ports
        ports = ChannelList(self, "PNAPorts", PNAPort)
        for port_num in range(1, nports + 1):
            port = PNAPort(
                self, f"port{port_num}", port_num, min_power, max_power
            )
            ports.append(port)
            self.add_submodule(f"port{port_num}", port)
        self.add_submodule("ports", ports.to_channel_tuple())

        # RF output
        self.add_parameter(
            "output",
            label="RF Output",
            get_cmd=":OUTPut?",
            set_cmd=":OUTPut {}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
        )

        # Drive power
        self.add_parameter(
            "power",
            label="Power",
            get_cmd="SOUR:POW?",
            get_parser=float,
            set_cmd="SOUR:POW {:.2f}",
            unit="dBm",
            vals=Numbers(min_value=min_power, max_value=max_power),
        )

        # IF bandwidth
        self.add_parameter(
            "if_bandwidth",
            label="IF Bandwidth",
            get_cmd="SENS:BAND?",
            get_parser=float,
            set_cmd="SENS:BAND {:.2f}",
            unit="Hz",
            vals=Numbers(min_value=1, max_value=15e6),
        )

        # Number of averages (also resets averages)
        self.add_parameter(
            "averages_enabled",
            label="Averages Enabled",
            get_cmd="SENS:AVER?",
            set_cmd="SENS:AVER {}",
            val_mapping={True: "1", False: "0"},
        )
        self.add_parameter(
            "averages",
            label="Averages",
            get_cmd="SENS:AVER:COUN?",
            get_parser=int,
            set_cmd="SENS:AVER:COUN {:d}",
            unit="",
            vals=Numbers(min_value=1, max_value=65536),
        )

        # Setting frequency range
        self.add_parameter(
            "start",
            label="Start Frequency",
            get_cmd="SENS:FREQ:STAR?",
            get_parser=float,
            set_cmd="SENS:FREQ:STAR {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )
        self.add_parameter(
            "stop",
            label="Stop Frequency",
            get_cmd="SENS:FREQ:STOP?",
            get_parser=float,
            set_cmd="SENS:FREQ:STOP {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )
        self.add_parameter(
            "center",
            label="Center Frequency",
            get_cmd="SENS:FREQ:CENT?",
            get_parser=float,
            set_cmd="SENS:FREQ:CENT {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )
        self.add_parameter(
            "span",
            label="Frequency Span",
            get_cmd="SENS:FREQ:SPAN?",
            get_parser=float,
            set_cmd="SENS:FREQ:SPAN {}",
            unit="Hz",
            vals=Numbers(min_value=0, max_value=max_freq),
        )
        self.add_parameter(
            "cw",
            label="CW Frequency",
            get_cmd="SENS:FREQ:CW?",
            get_parser=float,
            set_cmd="SENS:FREQ:CW {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )

        # Number of points in a sweep
        self.add_parameter(
            "points",
            label="Points",
            get_cmd="SENS:SWE:POIN?",
            get_parser=int,
            set_cmd="SENS:SWE:POIN {}",
            unit="",
            vals=Numbers(min_value=1, max_value=100001),
        )

        # Electrical delay
        self.add_parameter(
            "electrical_delay",
            label="Electrical Delay",
            get_cmd="CALC:CORR:EDEL:TIME?",
            get_parser=float,
            set_cmd="CALC:CORR:EDEL:TIME {:.6e}",
            unit="s",
            vals=Numbers(min_value=0, max_value=100000),
        )

        # Time gating
        self.add_parameter(
            "gating_enabled",
            label="Gating Enabled",
            get_cmd="CALC:FILT:TIME:STAT?",
            set_cmd="CALC:FILT:TIME:STAT {}",
            val_mapping={True: "1", False: "0"},
        )
        self.add_parameter(
            "gating_start",
            label="Gating Start time",
            get_cmd="CALC:FILT:TIME:STAR?",
            get_parser=float,
            set_cmd="CALC:FILT:TIME:STAR {:.4e}",
            unit="s",
            vals=Numbers(min_value=0, max_value=10e-6),
        )
        self.add_parameter(
            "gating_stop",
            label="Gating Stop time",
            get_cmd="CALC:FILT:TIME:STOP?",
            get_parser=float,
            set_cmd="CALC:FILT:TIME:STOP {:.4e}",
            unit="s",
            vals=Numbers(min_value=0, max_value=10e-6),
        )

        # Time domain
        self.add_parameter(
            "time_domain_enabled",
            label="Time-domain Enabled",
            get_cmd="CALC:TRAN:TIME:STAT?",
            set_cmd="CALC:TRAN:TIME:STAT {}",
            val_mapping={True: "1", False: "0"},
        )
        self.add_parameter(
            "start_time",
            label="Start Time",
            get_cmd="CALC:TRAN:TIME:STAR?",
            get_parser=float,
            set_cmd="CALC:TRAN:TIME:STAR {}",
            unit="s",
            vals=Numbers(min_value=0, max_value=2.5e-6),
        )
        self.add_parameter(
            "stop_time",
            label="Stop Time",
            get_cmd="CALC:TRAN:TIME:STOP?",
            get_parser=float,
            set_cmd="CALC:TRAN:TIME:STOP {}",
            unit="s",
            vals=Numbers(min_value=0, max_value=2.5e-6),
        )

        # Sweep Time
        self.add_parameter(
            "sweep_time",
            label="Time",
            get_cmd="SENS:SWE:TIME?",
            get_parser=float,
            unit="s",
            vals=Numbers(0, 1e6),
        )
        # Sweep Mode
        self.add_parameter(
            "sweep_mode",
            label="Mode",
            get_cmd="SENS:SWE:MODE?",
            set_cmd="SENS:SWE:MODE {}",
            vals=Enum("HOLD", "CONT", "GRO", "SING"),
        )
        # Sweep Type
        self.add_parameter(
            "sweep_type",
            label="Type",
            get_cmd="SENS:SWE:TYPE?",
            set_cmd="SENS:SWE:TYPE {}",
            vals=Enum("LIN", "LOG", "POW", "CW", "SEGM", "PHAS"),
        )

        # Group trigger count
        self.add_parameter(
            "group_trigger_count",
            get_cmd="SENS:SWE:GRO:COUN?",
            get_parser=int,
            set_cmd="SENS:SWE:GRO:COUN {}",
            vals=Ints(1, 2000000),
        )
        # Trigger Source
        self.add_parameter(
            "trigger_source",
            get_cmd="TRIG:SOUR?",
            set_cmd="TRIG:SOUR {}",
            vals=Enum("EXT", "IMM", "MAN"),
        )

        # Axis Parameters
        self.add_parameter(
            "frequency_axis",
            unit="Hz",
            label="Frequency",
            parameter_class=PNAAxisParameter,
            startparam=self.start,
            stopparam=self.stop,
            pointsparam=self.points,
            vals=Arrays(shape=(self.points,)),
        )
        self.add_parameter(
            "frequency_log_axis",
            unit="Hz",
            label="Frequency",
            parameter_class=PNALogAxisParamter,
            startparam=self.start,
            stopparam=self.stop,
            pointsparam=self.points,
            vals=Arrays(shape=(self.points,)),
        )
        self.add_parameter(
            "time_axis",
            unit="s",
            label="Time",
            parameter_class=PNATimeAxisParameter,
            startparam=None,
            stopparam=self.sweep_time,
            pointsparam=self.points,
            vals=Arrays(shape=(self.points,)),
        )
        self.add_parameter(
            "time_domain_axis",
            unit="s",
            label="Time",
            parameter_class=PNATimeDomainAxisParameter,
            startparam=self.start_time,
            stopparam=self.stop_time,
            pointsparam=self.points,
            vals=Arrays(shape=(self.points,)),
        )

        # Traces
        self.add_parameter(
            "active_trace",
            label="Active Trace",
            get_cmd="CALC:PAR:MNUM?",
            get_parser=int,
            set_cmd="CALC:PAR:SEL 'Meas{}'",
            vals=Numbers(min_value=1, max_value=24),
        )
        # Note: Traces will be accessed through the traces property which
        # updates the channellist to include only active trace numbers
        self._traces = ChannelList(self, "PNATraces", PNATrace)
        self.add_submodule("traces", self._traces)
        # Add shortcuts to first trace
        trace1 = self.traces[0]
        params = trace1.parameters
        if not isinstance(params, dict):
            raise RuntimeError(
                f"Expected trace.parameters to be a dict got "
                f"{type(params)}"
            )
        for param in params.values():
            self.parameters[param.name] = param
        # And also add a link to run sweep
        self.run_sweep = trace1.run_sweep
        # Set this trace to be the default (it's possible to end up in a
        # situation where no traces are selected, causing parameter snapshots
        # to fail)
        self.active_trace(trace1.trace_num)

        # Set auto_sweep parameter
        # If we want to return multiple traces per setpoint without sweeping
        # multiple times, we should set this to false
        self.add_parameter(
            "auto_sweep",
            label="Auto Sweep",
            set_cmd=None,
            get_cmd=None,
            vals=Bool(),
            initial_value=True,
        )

        # A default output format on initialisation
        self.write("FORM REAL,32")
        self.write("FORM:BORD NORM")

        self.connect_message()

    @property
    def traces(self) -> ChannelList:
        """
        Update channel list with active traces and return the new list
        """
        # Keep track of which trace was active before. This command may fail
        # if no traces were selected.
        try:
            active_trace = self.active_trace()
        except errors.VisaIOError as e:
            self.log.debug("Exception on querying active trace: %r", e)
            if e.error_code == constants.StatusCode.error_timeout:
                self.log.info("No active trace on PNA")
                active_trace = None
            else:
                raise

        # Get a list of traces from the instrument and fill in the traces list
        parlist = self.get_trace_catalog().split(",")
        self._traces.clear()
        for trace_name in parlist[::2]:
            trace_num = self.select_trace_by_name(trace_name)
            pna_trace = PNATrace(self, f"tr{trace_num}", trace_name, trace_num)
            self._traces.append(pna_trace)

        # Restore the active trace if there was one
        if active_trace:
            self.active_trace(active_trace)

        # Return the list of traces on the instrument
        return self._traces

    def get_options(self) -> Sequence[str]:
        # Query the instrument for what options are installed
        return self.ask("*OPT?").strip('"').split(",")

    def preset(self) -> None:
        self.write("SYST:FPRESET")
        self.write("DISP:WIND:STATE ON")

    def get_trace_catalog(self) -> str:
        """
        Get the trace catalog, that is a list of trace and sweep types
        from the PNA.
        The format of the returned trace is:
            trace_name,trace_type,trace_name,trace_type...
        """
        return self.ask("CALC:PAR:CAT:EXT?").strip('"')

    def select_trace_by_name(self, trace_name: str) -> int:
        """
        Select a trace on the PNA by name.
        Returns:
            The trace number of the selected trace
        """
        self.write(f"CALC:PAR:SEL '{trace_name}'")
        return self.active_trace()

    def add_trace(self, trace_num: int, trace_meas: str) -> None:
        """
        Add a trace to the channel.
        """
        self.write(f"CALC:PAR:DEF:EXT 'Meas{trace_num}',{trace_meas}")
        self.write(f"DISP:WIND:TRACe{trace_num}:FEED 'Meas{trace_num}'")

    def reset_averages(self) -> None:
        """
        Reset averaging
        """
        self.write("SENS:AVER:CLE")

    def averages_on(self) -> None:
        """
        Turn on trace averaging
        """
        self.averages_enabled(True)

    def averages_off(self) -> None:
        """
        Turn off trace averaging
        """
        self.averages_enabled(False)

    def gating_on(self) -> None:
        """
        Turn on trace gating
        """
        self.gating_enabled(True)

    def gating_off(self) -> None:
        """
        Turn off trace gating
        """
        self.gating_enabled(False)

    def time_domain_on(self) -> None:
        """
        Turn on trace transform to time-domain
        """
        self.time_domain_enabled(True)

    def time_domain_off(self) -> None:
        """
        Turn off trace transform to time-domain
        """
        self.time_domain_enabled(False)

    def autoscale(self) -> None:
        """
        Autoscale windows
        """
        self.write("DISP:WIND:Y:AUTO")

    def _set_power_limits(
        self, min_power: Union[int, float], max_power: Union[int, float]
    ) -> None:
        """
        Set port power limits
        """
        self.power.vals = Numbers(min_value=min_power, max_value=max_power)
        for port in self.ports:
            port._set_power_limits(min_power, max_power)


class PNABaseN52xx(VisaInstrument):
    """
    Base qcodes driver for Agilent/Keysight series PNAs
    http://na.support.keysight.com/pna/help/latest/Programming/GP-IB_Command_Finder/SCPI_Command_Tree.htm
    Note: Currently this driver only expects a single channel on the PNA. We
        can handle multiple traces, but using traces across multiple channels
        may have unexpected results.
    """

    def __init__(
        self,
        name: str,
        address: str,
        # Set frequency ranges
        min_freq: Union[int, float],
        max_freq: Union[int, float],
        # Set power ranges
        min_power: Union[int, float],
        max_power: Union[int, float],
        nports: int,  # Number of ports on the PNA
        **kwargs: Any,
    ) -> None:
        super().__init__(name, address, terminator="\n", **kwargs)
        self.min_freq = min_freq
        self.max_freq = max_freq

        self.log.info(
            "Initializing %s with power range %r-%r, freq range %r-%r.",
            name,
            min_power,
            max_power,
            min_freq,
            max_freq,
        )

        # Ports
        ports = ChannelList(self, "PNAPorts", PNAPort)
        for port_num in range(1, nports + 1):
            port = PNAPort(
                self, f"port{port_num}", port_num, min_power, max_power
            )
            ports.append(port)
            self.add_submodule(f"port{port_num}", port)
        self.add_submodule("ports", ports.to_channel_tuple())

        # RF output
        self.add_parameter(
            "output",
            label="RF Output",
            get_cmd=":OUTPut?",
            set_cmd=":OUTPut {}",
            val_mapping=create_on_off_val_mapping(on_val="1", off_val="0"),
        )

        # Drive power
        self.add_parameter(
            "power",
            label="Power",
            get_cmd="SOUR:POW?",
            get_parser=float,
            set_cmd="SOUR:POW {:.2f}",
            unit="dBm",
            vals=Numbers(min_value=min_power, max_value=max_power),
        )

        # IF bandwidth
        self.add_parameter(
            "if_bandwidth",
            label="IF Bandwidth",
            get_cmd="SENS:BAND?",
            get_parser=float,
            set_cmd="SENS:BAND {:.2f}",
            unit="Hz",
            vals=Numbers(min_value=1, max_value=15e6),
        )

        # Number of averages (also resets averages)
        self.add_parameter(
            "averages_enabled",
            label="Averages Enabled",
            get_cmd="SENS:AVER?",
            set_cmd="SENS:AVER {}",
            val_mapping={True: "1", False: "0"},
        )
        self.add_parameter(
            "averages",
            label="Averages",
            get_cmd="SENS:AVER:COUN?",
            get_parser=int,
            set_cmd="SENS:AVER:COUN {:d}",
            unit="",
            vals=Numbers(min_value=1, max_value=65536),
        )

        # Setting frequency range
        self.add_parameter(
            "start",
            label="Start Frequency",
            get_cmd="SENS:FREQ:STAR?",
            get_parser=float,
            set_cmd="SENS:FREQ:STAR {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )
        self.add_parameter(
            "stop",
            label="Stop Frequency",
            get_cmd="SENS:FREQ:STOP?",
            get_parser=float,
            set_cmd="SENS:FREQ:STOP {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )
        self.add_parameter(
            "center",
            label="Center Frequency",
            get_cmd="SENS:FREQ:CENT?",
            get_parser=float,
            set_cmd="SENS:FREQ:CENT {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )
        self.add_parameter(
            "span",
            label="Frequency Span",
            get_cmd="SENS:FREQ:SPAN?",
            get_parser=float,
            set_cmd="SENS:FREQ:SPAN {}",
            unit="Hz",
            vals=Numbers(min_value=0, max_value=max_freq),
        )
        self.add_parameter(
            "cw",
            label="CW Frequency",
            get_cmd="SENS:FREQ:CW?",
            get_parser=float,
            set_cmd="SENS:FREQ:CW {}",
            unit="Hz",
            vals=Numbers(min_value=min_freq, max_value=max_freq),
        )

        # Number of points in a sweep
        self.add_parameter(
            "points",
            label="Points",
            get_cmd="SENS:SWE:POIN?",
            get_parser=int,
            set_cmd="SENS:SWE:POIN {}",
            unit="",
            vals=Numbers(min_value=1, max_value=100001),
        )

        # Sweep Time
        self.add_parameter(
            "sweep_time",
            label="Time",
            get_cmd="SENS:SWE:TIME?",
            get_parser=float,
            unit="s",
            vals=Numbers(0, 1e6),
        )
        # Sweep Mode
        self.add_parameter(
            "sweep_mode",
            label="Mode",
            get_cmd="SENS:SWE:MODE?",
            set_cmd="SENS:SWE:MODE {}",
            vals=Enum("HOLD", "CONT", "GRO", "SING"),
        )
        # Sweep Type
        self.add_parameter(
            "sweep_type",
            label="Type",
            get_cmd="SENS:SWE:TYPE?",
            set_cmd="SENS:SWE:TYPE {}",
            vals=Enum("LIN", "LOG", "POW", "CW", "SEGM", "PHAS"),
        )

        # Group trigger count
        self.add_parameter(
            "group_trigger_count",
            get_cmd="SENS:SWE:GRO:COUN?",
            get_parser=int,
            set_cmd="SENS:SWE:GRO:COUN {}",
            vals=Ints(1, 2000000),
        )
        # Trigger Source
        self.add_parameter(
            "trigger_source",
            get_cmd="TRIG:SOUR?",
            set_cmd="TRIG:SOUR {}",
            vals=Enum("EXT", "IMM", "MAN"),
        )

        # Axis Parameters
        self.add_parameter(
            "frequency_axis",
            unit="Hz",
            label="Frequency",
            parameter_class=PNAAxisParameter,
            startparam=self.start,
            stopparam=self.stop,
            pointsparam=self.points,
            vals=Arrays(shape=(self.points,)),
        )
        self.add_parameter(
            "frequency_log_axis",
            unit="Hz",
            label="Frequency",
            parameter_class=PNALogAxisParamter,
            startparam=self.start,
            stopparam=self.stop,
            pointsparam=self.points,
            vals=Arrays(shape=(self.points,)),
        )

        # Traces
        self.add_parameter(
            "active_trace",
            label="Active Trace",
            get_cmd="CALC:PAR:MNUM?",
            get_parser=int,
            set_cmd="CALC:PAR:MNUM {}",
            vals=Numbers(min_value=1, max_value=24),
        )
        # Note: Traces will be accessed through the traces property which
        # updates the channellist to include only active trace numbers
        self._traces = ChannelList(self, "PNATraces", PNATrace)
        self.add_submodule("traces", self._traces)
        # Add shortcuts to first trace
        trace1 = self.traces[0]
        params = trace1.parameters
        if not isinstance(params, dict):
            raise RuntimeError(
                f"Expected trace.parameters to be a dict got "
                f"{type(params)}"
            )
        for param in params.values():
            self.parameters[param.name] = param
        # And also add a link to run sweep
        self.run_sweep = trace1.run_sweep
        # Set this trace to be the default (it's possible to end up in a
        # situation where no traces are selected, causing parameter snapshots
        # to fail)
        self.active_trace(trace1.trace_num)

        # Set auto_sweep parameter
        # If we want to return multiple traces per setpoint without sweeping
        # multiple times, we should set this to false
        self.add_parameter(
            "auto_sweep",
            label="Auto Sweep",
            set_cmd=None,
            get_cmd=None,
            vals=Bool(),
            initial_value=True,
        )

        # A default output format on initialisation
        self.write("FORM REAL,32")
        self.write("FORM:BORD NORM")

        self.connect_message()

    @property
    def traces(self) -> ChannelList:
        """
        Update channel list with active traces and return the new list
        """
        # Keep track of which trace was active before. This command may fail
        # if no traces were selected.
        try:
            active_trace = self.active_trace()
        except errors.VisaIOError as e:
            self.log.debug("Exception on querying active trace: %r", e)
            if e.error_code == constants.StatusCode.error_timeout:
                self.log.info("No active trace on PNA")
                active_trace = None
            else:
                raise

        # Get a list of traces from the instrument and fill in the traces list
        parlist = self.get_trace_catalog().split(",")
        self._traces.clear()
        for trace_name in parlist[::2]:
            trace_num = self.select_trace_by_name(trace_name)
            pna_trace = PNATrace(self, f"tr{trace_num}", trace_name, trace_num)
            self._traces.append(pna_trace)

        # Restore the active trace if there was one
        if active_trace:
            self.active_trace(active_trace)

        # Return the list of traces on the instrument
        return self._traces

    def get_options(self) -> Sequence[str]:
        # Query the instrument for what options are installed
        return self.ask("*OPT?").strip('"').split(",")

    def preset(self) -> None:
        self.write("SYST:FPRESET")
        self.write("DISP:WIND:STATE ON")

    def get_trace_catalog(self) -> str:
        """
        Get the trace catalog, that is a list of trace and sweep types
        from the PNA.
        The format of the returned trace is:
            trace_name,trace_type,trace_name,trace_type...
        """
        return self.ask("CALC:PAR:CAT:EXT?").strip('"')

    def select_trace_by_name(self, trace_name: str) -> int:
        """
        Select a trace on the PNA by name.
        Returns:
            The trace number of the selected trace
        """
        self.write(f"CALC:PAR:SEL '{trace_name}'")
        return self.active_trace()

    def add_trace(self, trace_num: int, trace_meas: str) -> None:
        """
        Add a trace to the channel.
        """
        self.write(f"CALC:PAR:DEF:EXT 'Meas{trace_num}',{trace_meas}")
        self.write(f"DISP:WIND:TRACe{trace_num}:FEED 'Meas{trace_num}'")

    def reset_averages(self) -> None:
        """
        Reset averaging
        """
        self.write("SENS:AVER:CLE")

    def averages_on(self) -> None:
        """
        Turn on trace averaging
        """
        self.averages_enabled(True)

    def averages_off(self) -> None:
        """
        Turn off trace averaging
        """
        self.averages_enabled(False)

    def autoscale(self) -> None:
        """
        Autoscale windows
        """
        self.write("DISP:WIND:Y:AUTO")

    def _set_power_limits(
        self, min_power: Union[int, float], max_power: Union[int, float]
    ) -> None:
        """
        Set port power limits
        """
        self.power.vals = Numbers(min_value=min_power, max_value=max_power)
        for port in self.ports:
            port._set_power_limits(min_power, max_power)


class P9375A(PNABase):
    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(
            name,
            address,
            min_freq=300e3,
            max_freq=26.5e9,
            min_power=-40,
            max_power=10,
            nports=2,
            **kwargs,
        )


class E8362B(PNABase):
    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(
            name,
            address,
            min_freq=10e6,
            max_freq=20e9,
            min_power=-40,
            max_power=13,
            nports=2,
            **kwargs,
        )


class N5224A(PNABaseN52xx):
    def __init__(self, name: str, address: str, **kwargs: Any):
        super().__init__(
            name,
            address,
            min_freq=10e6,
            max_freq=43.5e9,
            min_power=-30,
            max_power=13,  # 13 dBm up to 35 GHz
            nports=2,
            **kwargs,
        )



import hvplot
import plotly.graph_objects as go


def hv2plotly(plot, layout=None, trace_info=None, renderer="jupyterlab"):
    hvplot.output(backend="plotly")
    fig_hv = hvplot.render(plot)
    fig = go.Figure(fig_hv)
    default_layout = dict(
        template="simple_white",
        font=dict(family="Arial, monospace", size=30, color="Black"),
        margin=dict(l=100, r=25, t=25, b=100),
        xaxis=dict(
            ticks="inside",
            mirror="all",
            showgrid=True,
        ),
        yaxis=dict(
            ticks="inside",
            mirror="all",
            showgrid=True,
        ),
    )
    fig.update_layout(default_layout)
    if layout is not None:
        fig.update_layout(layout)
    if trace_info is not None:
        fig.update_traces(trace_info)
    if renderer is not None:
        fig.show(renderer=renderer)
    return fig



import time
from typing import Any

import nidaqmx
import numpy as np
import serial
from qcodes.instrument.base import Instrument
from qcodes.instrument.parameter import Parameter
from qcodes.validators import Arrays, Bool, Enum, Ints, Numbers

V_UNIT = nidaqmx.constants.VoltageUnits.VOLTS
V_IN_TERM_CONF = nidaqmx.constants.TerminalConfiguration.RSE
V_MIN = -10
V_MAX = 10


class OutputField(Parameter):
    """Output magnetic field by writing voltage value to daq.
    Args:
        name: Name of parameter.
        dev_name: DAQ device name (e.g. 'Dev1').
        v_out_address: AO address.
        time_voltage_sweep: DAQ AO time to sweep voltage.
        write_sample_clock: DAQ AO sampling clock.
        write_array_length: DAQ AO sampling length.
        kwargs: Keyword arguments to be passed to ArrayParameter constructor.
    """

    def __init__(
        self,
        name: str,
        dev_name: str,
        v_out_address: str,
        time_voltage_sweep: float,
        write_sample_clock: int,
        write_array_length: int,
        **kwargs,
    ) -> None:
        super().__init__(name, **kwargs)
        self.dev_name = dev_name
        self.v_out_address = v_out_address
        self.time_voltage_sweep = time_voltage_sweep
        self.write_sample_clock = write_sample_clock
        self.write_array_length = write_array_length
        self._voltage = 0
        self._field = 0

        task = nidaqmx.Task()
        task.ao_channels.add_ao_voltage_chan(
            f"{self.dev_name}/{self.v_out_address}",
            min_val=V_MIN,
            max_val=V_MAX,
        )
        task.timing.cfg_samp_clk_timing(self.write_sample_clock)
        v_out = np.zeros(int(self.write_array_length))
        task.write(v_out, auto_start=True)
        task.wait_until_done()
        task.stop()
        task.close()

    def set_raw(self, field: float) -> None:
        """Set output voltage of magnet"""
        voltage = self.H2V(field)
        task = nidaqmx.Task()
        task.ao_channels.add_ao_voltage_chan(
            f"{self.dev_name}/{self.v_out_address}",
            min_val=V_MIN,
            max_val=V_MAX,
        )
        task.timing.cfg_samp_clk_timing(self.write_sample_clock)
        v_out = np.linspace(
            self._voltage, voltage, int(self.write_array_length)
        )
        task.write(v_out, auto_start=True)
        task.wait_until_done()
        task.stop()
        task.close()
        self._voltage = voltage
        self._field = field
        time.sleep(0.1)
        return

    def get_raw(self):
        """Returns last voltage written to outputs."""
        return self._field

    def H2V(self, H: float) -> float:
        """Convert target magnetic field to output voltage.
        input: magnetic field (mT)
        output: voltage (V)
        """
        H = H * 10  # convert from mT to Oe
        V = (
            -0.01268
            + 0.00281 * H
            + 3.02608e-10 * H**2
            - 1.5036e-11 * H**3
            - 1.66895e-16 * H**4
            + 3.15376e-18 * H**5
        )
        return V


class MeasuredField(Parameter):
    """Measured magnetic field by reading voltage value to daq
    Args:
        name: Name of parameter.
        dev_name: DAQ device name (e.g. 'Dev1').
        v_hall_address: DAQ AI ch. of hall voltage.
        read_sample_num: DAQ AI sampling number.
        read_sample_clock: DAQ AI sampling clock.
        kwargs: Keyword arguments to be passed to ArrayParameter constructor.
    """

    def __init__(
        self,
        name: str,
        dev_name: str,
        v_hall_address: str,
        read_sample_num: int,
        read_sample_clock: int,
        **kwargs,
    ) -> None:
        super().__init__(name, **kwargs)
        self.dev_name = dev_name
        self.v_hall_address = v_hall_address
        self.read_sample_num = read_sample_num
        self.read_sample_clock = read_sample_clock
        self._voltage = np.nan
        self._field = np.nan

    def get_raw(self):
        """Returns measured magnetic field from hall voltage"""
        task = nidaqmx.Task()
        task.ai_channels.add_ai_voltage_chan(
            f"{self.dev_name}/{self.v_hall_address}",
            min_val=V_MIN,
            max_val=V_MAX,
            terminal_config=V_IN_TERM_CONF,
            units=V_UNIT,
        )
        task.timing.cfg_samp_clk_timing(self.read_sample_clock)
        data = task.read(number_of_samples_per_channel=self.read_sample_num)
        task.wait_until_done()
        task.stop()
        task.close()
        V = np.mean(data)
        self._voltage = V
        self._field = self.V2H(self._voltage)
        return self._field

    def V2H(self, V: float) -> float:
        """Convert hall voltage to magnetic field
        input: hall voltage (V)
        output: magnetic field (mT)
        """
        H = (
            -0.05665
            + 260.16273 * V
            - 2.1979e-5 * V**2
            + 0.01858 * V**3
            + 1.97225e-4 * V**4
            - 2.33726e-4 * V**5
        )
        H = H / 10  # convert from Oe to mT
        return H


class Magnet(Instrument):
    """
    This is the QCoDeS python driver to control the magnet of
    TOEI in-plane magnetic prober.
    """

    def __init__(
        self,
        name: str,
        dev_name: str,
        v_hall_address: str,
        v_out_address: str,
        read_sample_num: int = 100,
        read_sample_clock: int = 10000,
        time_voltage_sweep: float = 10e-3,
        write_sample_clock: int = 100e3,
        write_array_length: int = 1e3,
        **kwargs,
    ) -> None:
        """
        QCoDeS driver for TOEI in-plane magnetic prober.
        Args:
            name: Name of the instrument.
            dev_name: DAQ device name (e.g. 'Dev1').
            v_hall_address: DAQ AI ch. of hall voltage.
            v_out_address: DAQ AO ch. of magnet voltage.
            read_sample_num: DAQ AI sampling number.
            read_sample_clock: DAQ AI sampling clock.
            time_voltage_sweep: DAQ AO time to sweep voltage.
            write_sample_clock: DAQ AO sampling clock.
            write_array_length: DAQ AO sampling length.
        """

        super().__init__(name=name, **kwargs)
        self.metadata.update(
            {
                "dev_name": dev_name,
                "v_hall_address": v_hall_address,
                "v_out_address": v_out_address,
                "read_sample_num": read_sample_num,
                "read_sample_clock": read_sample_clock,
                "time_voltage_sweep": time_voltage_sweep,
                "write_sample_clock": write_sample_clock,
                "write_array_length": write_array_length,
            }
        )

        self.add_parameter(
            name="measured_field",
            dev_name=dev_name,
            v_hall_address=v_hall_address,
            read_sample_num=read_sample_num,
            read_sample_clock=read_sample_clock,
            parameter_class=MeasuredField,
            label="Magnetic field",
            unit="mT",
        )
        self.add_parameter(
            name="output_field",
            dev_name=dev_name,
            v_out_address=v_out_address,
            time_voltage_sweep=time_voltage_sweep,
            write_sample_clock=write_sample_clock,
            write_array_length=write_array_length,
            parameter_class=OutputField,
            label="Magnetic field",
            unit="mT",
        )

        self.connect_message()

    def get_idn(self):
        return {
            "vendor": "TOEI",
            "model": "In-plane prober",
            "serial": None,
            "firmware": None,
        }


class Angle(Parameter):
    """Measured magnetic field by reading voltage value to daq
    Args:
        name: Name of parameter.
        instrument: AngleDriver class.
        kwargs: Keyword arguments to be passed to ArrayParameter constructor.
    """

    def __init__(
        self,
        name: str,
        instrument: "AngleDriver",
        **kwargs,
    ) -> None:
        super().__init__(name, instrument=instrument, **kwargs)

    def set_raw(self, angle: float) -> None:
        """Sets magnetic field angle"""
        position = int(
            angle * self.instrument.root_instrument.angle_length_ratio
        )
        self.instrument.root_instrument.write(f"FP{position}")
        while True:
            motor_status = self.instrument.root_instrument.ask("RS")
            if motor_status == "R":
                break
            elif motor_status == "FMR":
                time.sleep(1)

    def get_raw(self) -> float:
        """Returns magnetic field angle"""
        position = self.instrument.root_instrument.ask("IP")
        position = float(position)
        angle = position / self.instrument.root_instrument.angle_length_ratio
        return angle


class AngleDriver(Instrument):
    """
    Driver for the motor which rotates the magnet
    """

    def __init__(
        self,
        name: str,
        com_port: str,
        baudrate: int = 9600,
        acceleration: float = 0.167,
        deceleration: float = 0.167,
        velocity: float = 0.05,
        step_resolution: int = 20000,
        angle_length_ratio: float = 22600 / 90,
        **kwargs,
    ) -> None:

        super().__init__(name=name, **kwargs)
        self.angle_length_ratio = angle_length_ratio
        self.metadata.update(
            {
                "com_port": com_port,
                "baudrate": baudrate,
                "acceleration": acceleration,
                "deceleration": deceleration,
                "velocity": velocity,
                "step_resolution": step_resolution,
                "angle_length_ratio": angle_length_ratio,
            }
        )
        try:
            self.ser = serial.Serial()
            self.ser.port = com_port
            self.ser.baudrate = baudrate
            self.ser.bytesize = serial.EIGHTBITS
            self.ser.parity = serial.PARITY_NONE
            self.ser.stopbits = serial.STOPBITS_ONE
            self.ser.timeout = 0.1
            self.ser.xonxoff = False
            self.ser.rtscts = False
            self.ser.dsrdtr = False
            self.ser.writeTimeout = 0

            self.ser.open()
            time.sleep(1)
        except Exception as e:
            print("Error opening serial port")
            print(e)
            exit()

        if self.ser.isOpen():
            try:
                self.ser.flushInput()
                self.ser.flushOutput()
                self.write(
                    f"EG{step_resolution}"
                )  # Sets microstepping to 20,000 steps per revolution
                self.write(
                    "IFD"
                )  # Sets the format of drive responses to decimal
                self.write("SP0")  # Sets the starting position at 0
                self.write("AR")  # Alarm reset
                self.write(f"AC{acceleration}")  # Acceleration
                self.write(f"DE{deceleration}")  # Deceleration
                self.write(f"VE{velocity}")  # Velocity
                self.write("ME")  # Enable Motor
            except Exception as e1:
                print("Error Communicating...: " + str(e1))
        else:
            print("Cannot open serial port ")

        self.add_parameter(
            name="angle",
            parameter_class=Angle,
            label="Magnetic field angle",
            unit="degree",
        )

        self.connect_message()

    def get_idn(self):
        return {
            "vendor": "Applied Motion",
            "model": "ST5-SI",
            "serial": None,
            "firmware": None,
        }

    def write(self, cmd: str) -> None:
        self.ser.write((cmd + "\r").encode())
        response = self.ser.read(15).decode()
        if len(response) > 0:
            self.ser.flushInput()

    def ask(self, cmd: str) -> Any:
        self.ser.write((cmd + "\r").encode())
        response = self.ser.read(15).decode()
        if len(response) > 0:
            res_value = str.strip(response)[3:]
            self.ser.flushInput()
        else:
            res_value = None
        return res_value

    def homing(self, current_angle) -> None:
        home_position = int(-current_angle * self.angle_length_ratio)
        self.write(f"FP{home_position}")
        self.write("SP0")  # Sets the starting position at 0



from typing import Any
from qcodes.instrument import VisaInstrument
from qcodes.parameters import Parameter
from qcodes.validators import Enum, Strings

class KeysightDSOX2014A(VisaInstrument):
    """
    This is the QCoDeS driver for the Keysight_DSOX2014A Oscilloscope.
    """

    def __init__(self, name: str, address: str, channel: int, **kwargs: Any) -> None:
        super().__init__(name, address, terminator="\n", **kwargs)

        idn = self.IDN.get()
        self.model = idn["model"]

        self.add_parameter(
            "voltage",
            get_cmd=f"MEAS:VAV? CHAN{channel}",
            label="Voltage",
            get_parser=float,
            unit="V",
        )
        self.connect_message()

    def reset(self) -> None:
        self.write("*RST")